package de.korap.json;

import java.io.PrintStream;
import java.util.Iterator;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;

public class JsonTraverse

{
	// kind of operations during JsonTree Traversal:
	
	public final static int COLLECT 	= 1; 
	public final static int REWRITE 	= 2;	// replace lemmata by refs on expansion list.
	public final static int REWRITE_EXP	= 3;	// replace lemmata by expansion list [wf,wf,...].
	
	PrintStream
		fout,			// to be set by caller, replaces stdout.
		ferr;			// to be set by caller, replaces stderr.
	
	/*
	 * traverseJsonTree
	 * 
	 * - traverses JsonTree starting at node.
	 * - tPath: collects lemmata and contains references to rewrite.
	 * - mode: operation mode COLLECT, REWRITE, REWRITE_EXP.
	 * 31.03.20/FB
	 */
	
	static final public void traverseJsonTree(
				JsonNode 			node, 
				TraversedPath	 	tPath, 
				final int	 		mode, 
				final PrintStream	fout,
				final PrintStream	ferr )
	
	{
	final String func = "traverseJsonTree";
	boolean
		isFound = false;
	boolean
		bDebug = false;
	
    if( node.isObject())
    	{
        Iterator<String> fieldNames = node.fieldNames();
        if( bDebug )
        	fout.printf("Debug:%s: object: { \n",  func);
        
        while(fieldNames.hasNext()) 
        	{
            String fieldName = fieldNames.next();
            
            // path search under 'query':
            if( tPath.checkPath(fieldName, node.get(fieldName), mode, fout) )
            	isFound = true;
            
            JsonNode fieldValue = node.get(fieldName);
            if( bDebug )
            	fout.printf("Debug:%s: key='%s':\n", func, fieldName); 
            traverseJsonTree(fieldValue, tPath, mode, fout, ferr);
        	}
        
        if( bDebug )
        	fout.printf("Debug:%s: } : end of object.\n",  func);
        
        if( (mode == REWRITE || mode == REWRITE_EXP) && isFound )
        	{
        	tPath.rewriteObjNode(node, mode, fout);
        	fout.printf("Debug:%s: after rewrite: node='%s'.\n", func, node.toString());
        	}
    	} 
    else if(node.isArray())
    	{
    	if( bDebug )
    		fout.printf("Debug:%s: array: [\n",  func);
    	
        ArrayNode arrayNode = (ArrayNode) node;
        for(int i = 0; i < arrayNode.size(); i++) 
        	{
            JsonNode arrayElement = arrayNode.get(i);
            traverseJsonTree(arrayElement, tPath, mode, fout, ferr);
        	}
    	if( bDebug ) 
    		fout.printf("Debug:%s: ] end of array.\n",  func);
    	} 
    else {
        // JsonNode node represents a single value field - do something with it.
    	if( bDebug )
    		fout.printf("Debug:%s: single value field = '%s'\n", func, node.asText());
    	}
    
	} // traverseJsonTree
	
	/*
	 * isPathLemmaQuery
	 * 
	 * returns: lemma as a String, if path & lemma were found.
	 * 			else: null.
	 * 04.04.20/FB
	 */
	
	static private String isPathLemmaQuery(JsonNode node, PrintStream fout, PrintStream ferr)
	
	{
	final String func = "isPathLemmaQuery";
	int
		nMatches = 0;
	String
		lemma = null;
	boolean
		bDebug = false;
	
	if( node.isObject() == false )
		return null;
	
	Iterator<String> fieldNames = node.fieldNames();
    
    while(fieldNames.hasNext()) 
    	{
        String fieldName = fieldNames.next();
        String fieldValue = node.get(fieldName).asText();
        
        if( fieldName.equals("@type") && fieldValue.equals("koral:token") )
        	nMatches++;
        else if( fieldName.equals("wrap") && node.get(fieldName).isObject() == true ) 
	        {
        	Iterator<String>fieldNames2 = node.get(fieldName).fieldNames();
        	while( fieldNames2.hasNext())
	        	{
        		String fn = fieldNames2.next();
        		if( fn.equals("@type") && node.get(fn).asText().equals("koral:term") )
        			nMatches++;
        		else if( fn.equals("key") )
	        		{
	        		lemma = node.get(fn).asText();	
		        	}
           		else if( fn.equals("layer") && node.get(fn).asText().equals("lemma") )
	        		nMatches++;
	        	} // while
	        }
    	} // while
	
    if( bDebug )
    	fout.printf("Debug: %s: nMatches=%d.\n", func, nMatches);
    
    if( nMatches == 3 && lemma != null )
	    {
	    fout.printf("Debug:%s: path found to lemma '%s'.\n", func, lemma);
	    return lemma;
	    }
	
    // path of lemma not found:
	return null;
	} // isPathLemmaQuery
	
} // JsonTraverse
